import sys
import argparse

from typing import TYPE_CHECKING, Any

from ..common import PRESETS

if TYPE_CHECKING:
    SubArgumentParser = argparse._SubParsersAction[argparse.ArgumentParser]
else:
    SubArgumentParser = Any


def add_config_cli(parser: SubArgumentParser) -> None:
    ap = parser.add_parser(
        "config",
        description="Generate template configfile.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    ap.add_argument(
        "-x",
        "--preset",
        type=str,
        default=None,
        choices=PRESETS,
        help="Configuration preset to use.",
    )
    ap.add_argument(
        "-o",
        "--outfile",
        type=argparse.FileType("wt"),
        default=sys.stdout,
        help="Output config file.",
    )

    return None



import io
import os
import gzip
import logging
from collections import OrderedDict, defaultdict
from typing import DefaultDict, Generator, Iterable, TextIO

import pyBigWig  # type: ignore
import numpy as np
import polars as pl
from matplotlib.colors import rgb2hex
from intervaltree import Interval  # type: ignore[import-untyped]

from ..common import BED9P_COLS, STATUSES, add_group_columns
from .region import Action, ActionOpt, Region

logger = logging.getLogger(__name__)


IDENT_BREAKPOINTS = [
    85.0,
    90.0,
    95.0,
    97.5,
    98.0,
    98.5,
    98.75,
    99.0,
    99.25,
    99.5,
    99.75,
    100.0,
]
IDENT_COLORS = [
    "#4b3991",
    "#2974af",
    "#4a9da8",
    "#57b894",
    "#9dd893",
    "#e1f686",
    "#ffffb2",
    "#fdda79",
    "#fb9e4f",
    "#ee5634",
    "#c9273e",
    "#8a0033",
]


def read_bed_file(
    bed_file: TextIO,
) -> Generator[tuple[str, int, int, list[str]], None, None]:
    for line in bed_file.readlines():
        if line[0] == "#":
            continue
        chrm, start, end, *other = line.strip().split("\t")
        yield (chrm, int(start), int(end), other)


def read_regions(bed_file: TextIO, action: ActionOpt) -> Generator[Region, None, None]:
    for i, line in enumerate(read_bed_file(bed_file)):
        ctg, start, end, other = line
        try:
            desc = other[0]
        except IndexError:
            desc = None
        try:
            item_rgb = other[5]
        except IndexError:
            item_rgb = None

        # Allow item_rgb to be hexcode or not.
        if action == ActionOpt.PLOT:
            if item_rgb and not item_rgb.startswith("#"):
                rgb = tuple(int(v) / 255 for v in item_rgb.split(","))
                assert len(rgb) == 3, "Invalid RGB value."
                action_desc = rgb2hex(rgb)
            else:
                action_desc = item_rgb
        else:
            action_desc = None

        yield Region(
            name=ctg,
            region=Interval(start, end),
            desc=desc,
            action=Action(action, action_desc),
        )


def read_ignored_regions(infile: TextIO) -> DefaultDict[str, set[Region]]:
    ignored_regions: DefaultDict[str, set[Region]] = defaultdict(set)
    for region in read_regions(infile, action=ActionOpt.IGNORE):
        ignored_regions[region.name].add(region)

    return ignored_regions


def read_overlay_regions(
    infiles: Iterable[TextIO],
) -> DefaultDict[str, OrderedDict[str, set[Region]]]:
    """
    Read input overlay BED files.
    """
    overlay_regions: defaultdict[str, OrderedDict[str, set[Region]]] = defaultdict(
        OrderedDict
    )
    # 1 as zero idx reserved for mapq
    for i, bed in enumerate(infiles, 1):
        track_label = f"Track {i}"
        for region in read_regions(bed, action=ActionOpt.PLOT):
            if track_label not in overlay_regions[region.name]:
                overlay_regions[region.name][track_label] = set()

            overlay_regions[region.name][track_label].add(region)

    return overlay_regions


def write_bigwig(
    df_pileup: pl.DataFrame, chrom_lengths: str, columns: list[str], output_dir: str
):
    chrom = df_pileup["chrom"][0]
    start = max(0, df_pileup["pos"][0] - 1)
    end = df_pileup["pos"][-1]
    chrom_coords = f"{chrom}_{start}-{end}"

    if not os.path.exists(chrom_lengths):
        logging.warning(
            f"Chromosome lengths (-f) are required to generate bigWig files for {chrom}. Generating wig files."
        )
        for col in columns:
            outfile = os.path.join(output_dir, f"{chrom_coords}_{col}.wig.gz")
            col = "bin_ident" if col == "ident" else col
            with gzip.open(outfile, "wb") as fh:
                df_values = (
                    df_pileup.select(col)
                    .cast({col: pl.Float64})
                    .rename({col: f"fixedStep chrom={chrom} start={start} step=1"})
                )
                df_values.write_csv(fh, include_header=True)  # type: ignore[call-overload]
    else:
        df_chrom_lengths = pl.read_csv(
            chrom_lengths,
            separator="\t",
            has_header=False,
            new_columns=["chrom", "length"],
            columns=[0, 1],
        ).filter(pl.col("chrom") == chrom)

        header = list(df_chrom_lengths.iter_rows())
        for col in columns:
            outfile = os.path.join(output_dir, f"{chrom_coords}_{col}.bw")
            col = "bin_ident" if col == "ident" else col
            with pyBigWig.open(outfile, "w") as bw:
                # https://github.com/deeptools/pyBigWig/issues/126
                bw.addHeader(header)
                values = df_pileup[col].to_numpy().astype(np.float64)
                bw.addEntries(chrom, start, values=values, span=1, step=1)


def generate_status_from_regions(df_region: pl.DataFrame) -> pl.DataFrame:
    # Regions don't have coordinates so need to group by break in contiguity of adjacent intervals.
    df_region_grp = add_group_columns(df_region)
    return (
        df_region_grp.group_by(["#chrom", "name", "group"])
        .agg(
            chromStart=pl.col("minStart").first(),
            chromEnd=pl.col("maxEnd").first(),
            perc=(
                pl.col("length").sum()
                / (pl.col("maxEnd").first() - pl.col("minStart").first())
            )
            * 100.0,
        )
        .pivot(
            on="name",
            index=["#chrom", "chromStart", "chromEnd"],
            values="perc",
            maintain_order=True,
        )
        # Ensure column exists.
        # https://github.com/pola-rs/polars/issues/18372#issuecomment-2390371173
        .with_columns(
            **{status: pl.coalesce(pl.col(f"^{status}$"), 0.0) for status in STATUSES},
        )
        .with_columns(
            status=pl.when(pl.col("correct") == 100.0)
            .then(pl.lit("correct"))
            .otherwise(pl.lit("misassembled")),
        )
        .select(
            pl.col("#chrom"),
            pl.col("chromStart"),
            pl.col("chromEnd"),
            pl.col("status"),
            *[pl.col(status) for status in STATUSES],
        )
        .fill_null(0.0)
        # chromStart and chromEnd get cast to float for some reason.
        .cast({"chromStart": pl.Int64, "chromEnd": pl.Int64})
        .sort(by=["#chrom", "chromStart"])
    )


def write_output(
    output_regions: TextIO | list[pl.DataFrame],
    output_status: TextIO | None,
) -> None:
    # If written to stdout, read saved inputs.
    if isinstance(output_regions, list):
        try:
            df_region = pl.concat(output_regions)
        except Exception:
            # No assembly errors.
            df_region = pl.DataFrame(schema=BED9P_COLS)
    elif isinstance(output_regions, io.TextIOBase):
        # Load out-of-order file.
        try:
            df_region = pl.read_csv(
                output_regions.name,
                has_header=False,
                separator="\t",
                schema=dict(BED9P_COLS),
                raise_if_empty=False,
                truncate_ragged_lines=True,
            )
            # Erase file and then rewrite in sorted order.
            output_regions.seek(0)
            output_regions.truncate(0)
            df_region.unique().sort(by=["#chrom", "chromStart"]).write_csv(
                file=output_regions, include_header=True, separator="\t"
            )
        except pl.exceptions.ShapeError:
            df_region = pl.DataFrame(schema=BED9P_COLS)
        except FileNotFoundError:
            return
    else:
        raise ValueError(f"Invalid misasm output. {output_regions}")
    if not output_status:
        return

    df_status = generate_status_from_regions(df_region)
    df_status.write_csv(file=output_status, include_header=True, separator="\t")


def read_identity_breakpoints(infile: TextIO | None) -> tuple[list[float], list[str]]:
    if not infile:
        return IDENT_BREAKPOINTS, IDENT_COLORS

    idents, colors = [], []
    for line in infile:
        ident, hexcode_color = line.strip().split("\t")
        idents.append(float(ident))
        colors.append(hexcode_color)

    return idents, colors



import sys
import ast
import argparse

from typing import TYPE_CHECKING, Any

from ..common import BED9P_COLS, STATUSES, PRESETS

if TYPE_CHECKING:
    SubArgumentParser = argparse._SubParsersAction[argparse.ArgumentParser]
else:
    SubArgumentParser = Any


def add_call_cli(parser: SubArgumentParser) -> None:
    ap = parser.add_parser(
        "call",
        description="Use per-base read coverage to classify/plot misassemblies",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    input_args = ap.add_argument_group(title="Input", description="Input files.")
    input_args.add_argument(
        "-i",
        "--infile",
        required=True,
        help="Indexed BAM or CRAM file.",
    )
    input_args.add_argument(
        "-f",
        "--fasta",
        default=None,
        help="Reference fasta. Used to bin pileup using average nucleotide identity and detect repeats.",
    )
    input_args.add_argument(
        "-b",
        "--input_regions",
        default=None,
        type=argparse.FileType("rt"),
        help="BED file with regions to check.",
    )
    input_args.add_argument(
        "--ignore_regions",
        default=None,
        type=argparse.FileType("rt"),
        help="Bed file with regions to ignore. With format: [contig, start, end]",
    )
    output_args = ap.add_argument_group(title="Outputs", description="Output files.")
    output_args.add_argument(
        "-o",
        "--output_regions",
        default=sys.stdout,
        type=argparse.FileType("wt"),
        help=f"Output bed file with checked regions. With format: {[c[0] for c in BED9P_COLS]}",
    )
    output_args.add_argument(
        "-s",
        "--output_status",
        default=None,
        type=argparse.FileType("wt"),
        help="Bed file with status of contigs and percentage breakdown of each misassembly type.",
    )
    output_args.add_argument(
        "-d",
        "--output_plot_dir",
        default=None,
        help="Output plot dir.",
    )
    output_args.add_argument(
        "--output_pileup_dir",
        default=None,
        help="Output pileup dir. Generates bigWig files per region.",
    )
    output_args.add_argument(
        "--add_pileup_data",
        nargs="*",
        choices=[
            "cov",
            "mismatch",
            "mapq",
            "insertion",
            "deletion",
            "softclip",
            "ident",
        ],
        default=["cov", "mismatch"],
        help="Add these pileup data types as bigWigs to --output_pileup_dir.",
    )
    config_args = ap.add_argument_group(title="Config", description="Configuration.")
    config_args.add_argument(
        "-t",
        "--threads",
        default=2,
        type=int,
        help="Threads for nucflag classification.",
    )
    config_args.add_argument(
        "-p",
        "--processes",
        default=8,
        type=int,
        help="Processes for plotting.",
    )
    config_args.add_argument(
        "-x",
        "--preset",
        default=None,
        choices=PRESETS,
        help="Sequencing data specific preset.",
    )
    config_args.add_argument(
        "-c",
        "--config",
        default=None,
        type=str,
        help="Threshold/params as toml file.",
    )
    config_args.add_argument(
        "--ignore_mtypes",
        nargs="*",
        choices=[status for status in STATUSES if status != "correct"],
        help="Ignore call types from plot and output bedfile.",
    )
    plot_args = ap.add_argument_group(title="Plot", description="Plot arguments.")
    plot_args.add_argument(
        "--tracks",
        nargs="*",
        type=argparse.FileType("rt"),
        help="Add additional regions as BED4 or BED9 on top of coverage plot.",
    )
    plot_args.add_argument(
        "--overlap_tracks",
        nargs="*",
        type=argparse.FileType("rt"),
        help="Add additional regions as BED4 or BED9 overlapping coverage plot.",
    )
    plot_args.add_argument(
        "--add_builtin_tracks",
        nargs="*",
        type=str,
        choices=["mapq", "ident"],
        help="Add built-in tracks used in nucflag as overlay tracks.",
    )
    plot_args.add_argument(
        "--ident_breakpoints",
        type=argparse.FileType("rt"),
        default=None,
        help="Use TSV of breakpoint and hexcode color for identity track.",
    )
    plot_args.add_argument(
        "--overlap_calls",
        dest="overlap_calls",
        action="store_true",
        help="Overlap NucFlag calls with coverage plot instead of adding as separate track. Similar to original NucFlag < v1.0. Correct regions are not displayed.",
    )
    plot_args.add_argument(
        "--ylim",
        default=3.0,
        type=ast.literal_eval,
        help="Plot y-axis limit. If float, used as a scaling factor from mean. (ex. 3.0 is mean times 3)",
    )
    return None


def add_status_cli(parser: SubArgumentParser) -> None:
    ap = parser.add_parser(
        "status",
        description="Generate status output from misassembly calls.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    ap.add_argument(
        "-i",
        "--infile",
        type=argparse.FileType("rb"),
        required=True,
        help="Input NucFlag misassembly calls as BED9.",
    )
    ap.add_argument(
        "-o",
        "--outfile",
        default=sys.stdout,
        type=argparse.FileType("wt"),
        help="Bed file with status of contigs and percentage breakdown of each misassembly type.",
    )
    return None



import argparse

from typing import TYPE_CHECKING, Any


if TYPE_CHECKING:
    SubArgumentParser = argparse._SubParsersAction[argparse.ArgumentParser]
else:
    SubArgumentParser = Any


def add_breakdown_cli(parser: SubArgumentParser) -> None:
    ap = parser.add_parser(
        "breakdown",
        description="Generate bar plot breakdown of whole-genome NucFlag misassemblies",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    ap.add_argument(
        "-i",
        "--infile",
        required=True,
        type=argparse.FileType("rb"),
        help="Input whole-genome NucFlag misassembly calls as BED9.",
    )
    ap.add_argument(
        "-l",
        "--filter_length",
        type=int,
        default=10_000_000,
        help="Only plot contigs at least this long.",
    )
    ap.add_argument(
        "-t",
        "--type",
        default="percent",
        choices=["percent", "length"],
        help="Type of plot. Either percent or length breakdown.",
    )
    ap.add_argument(
        "-o", "--output_prefix", default="breakdown", help="Output file prefix."
    )

    return None



import sys
import argparse

from typing import TYPE_CHECKING, Any

from .constants import GOOD_REGIONS

if TYPE_CHECKING:
    SubArgumentParser = argparse._SubParsersAction[argparse.ArgumentParser]
else:
    SubArgumentParser = Any


def add_consensus_cli(parser: SubArgumentParser) -> None:
    ap = parser.add_parser(
        "consensus",
        description="Generate consensus from putative misassembly bedfiles.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    ap.add_argument(
        "-i",
        "--infiles",
        required=True,
        nargs="+",
        type=argparse.FileType("rb"),
        help=f"Input misassembly calls as BED4+. Any value in the fourth column is expected to be an error with the exceptions of: {GOOD_REGIONS}",
    )
    ap.add_argument(
        "-n",
        "--total_number_ovl",
        type=int,
        default=2,
        help="Number of overlaps across all --infiles to be considered true.",
    )
    ap.add_argument(
        "-p",
        "--perc_ovl",
        type=float,
        default=None,
        help="Required overlap for each call as percentage (0.8 => 80 percent overlap) to be considered true. Provide value for stricter callsets. If None, any overlap is considered true.",
    )
    ap.add_argument(
        "-t",
        "--type_ovl",
        help="Which type of overlap. Similar to bedtools: -f (Fraction of A), -F (Fraction of B), r (Reciprocal fraction A_B).",
        default="A",
        type=str,
        choices=["A", "B", "A_B"],
    )
    ap.add_argument(
        "-g",
        "--ignore_calls",
        type=str,
        nargs="*",
        help=f"Calls to ignore. Typically good calls: {GOOD_REGIONS}.",
    )
    ap.add_argument(
        "-d",
        "--dst",
        type=int,
        help="Distance to allow in overlap. Added to both ends of interval.",
        default=1,
    )
    ap.add_argument(
        "-o",
        "--outfile",
        type=argparse.FileType("wt"),
        default=sys.stdout,
        help="Output file. Produces concensus BED4 file with overlaps delimited by ',' in 4th column.",
    )

    return None



import argparse
from py_nucflag import get_config_from_preset  # type: ignore[import-untyped]


def get_config(args: argparse.Namespace) -> int:
    cfg = get_config_from_preset(args.preset)
    print(cfg, file=args.outfile)
    return 0



import logging
import argparse
import polars as pl
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patheffects as pe

from matplotlib.axes import Axes
from matplotlib.colors import rgb2hex
from matplotlib.patches import Patch


from ..common import BED9P_COLS, STATUSES, minimalize_ax


logger = logging.getLogger(__name__)


def create_breakdown_plot(args: argparse.Namespace) -> int:
    df_calls = pl.read_csv(
        args.infile,
        separator="\t",
        has_header=False,
        comment_prefix="#",
        columns=list(range(9)),
        schema=dict(BED9P_COLS[0:9]),
        truncate_ragged_lines=True,
    ).with_columns(length=pl.col("chromEnd") - pl.col("chromStart"))

    df_fai = df_calls.group_by(["#chrom"]).agg(
        length=pl.col("chromEnd").max() - pl.col("chromStart").min()
    )
    color_key = {
        name: rgb2hex([int(e) / 255.0 for e in itemRgb.split(",")])
        if not itemRgb.startswith("#")
        else itemRgb
        for name, itemRgb in df_calls.select("name", "itemRgb").unique().iter_rows()
    }
    chrom_names = (
        df_fai.filter(pl.col("length") > args.filter_length)
        .sort(by="length", descending=True)["#chrom"]
        .unique(maintain_order=True)
    )
    if chrom_names.is_empty():
        return 1

    df_fai = df_fai.filter(pl.col("#chrom").is_in(chrom_names)).cast(
        {"#chrom": pl.Enum(chrom_names)}
    )
    logger.info(f"Plotting {len(chrom_names)} chromosomes.")

    df_grouped_calls = (
        df_calls.filter(pl.col("#chrom").is_in(chrom_names))
        .group_by(["#chrom", "name"])
        .agg(pl.col("length").sum())
        .pivot(
            on="name",
            index="#chrom",
            values="length",
            maintain_order=True,
        )
        .with_columns(
            # Ensure column exists.
            # https://github.com/pola-rs/polars/issues/18372#issuecomment-2390371173
            **{status: pl.coalesce(pl.col(f"^{status}$"), 0.0) for status in STATUSES},
        )
        # TODO: There's got to be a better way to do this.
        .unpivot(index="#chrom", value_name="length", variable_name="name")
        .cast({"name": pl.Enum(STATUSES), "#chrom": pl.Enum(chrom_names)})
        .sort(by=["#chrom", "name"])
    )
    # Scale fig width by number of chroms
    fig, axes = plt.subplots(
        nrows=2,
        height_ratios=[0.9, 0.1],
        figsize=(len(chrom_names), 20),
        layout="constrained",
    )
    ax: Axes = axes[0]
    bottom = np.zeros(len(chrom_names))

    if args.type == "percent":
        ax.set_ylabel("Percent (%)")
        df_grouped_calls = (
            df_grouped_calls.join(df_fai, on="#chrom", how="left")
            .with_columns(length=(pl.col("length") / pl.col("length_right")) * 100.0)
            .drop("length_right")
        )
    else:
        ax.set_ylabel("Length (Mbp)")
        df_grouped_calls = df_grouped_calls.with_columns(pl.col("length") / 1_000_000)

    # https://matplotlib.org/stable/gallery/lines_bars_and_markers/bar_label_demo.html#bar-chart-with-labels
    for group, df_group in df_grouped_calls.group_by(["name"], maintain_order=True):
        group = group[0]
        values = df_group["length"].to_numpy()
        color = color_key.get(group)
        p = ax.bar(chrom_names, values, 0.6, label=group, bottom=bottom, color=color)
        bottom += values
        text_color = "#000000"
        ax.bar_label(
            p,
            label_type="center",
            fmt="{:,.1f}",
            color=text_color,
            path_effects=[pe.withStroke(linewidth=4, foreground="white")],
        )

    ax.set_title("Misassemblies by contig", loc="left", pad=20)
    ax.set_xticklabels(
        ax.get_xticklabels(), rotation=45, ha="right", rotation_mode="anchor"
    )

    for spine in ["top", "right"]:
        ax.spines[spine].set_visible(False)

    # Add legend.
    legend_ax: Axes = axes[1]
    minimalize_ax(legend_ax, remove_ticks=True)
    legend_ax.legend(
        handles=[Patch(facecolor=color, label=lbl) for lbl, color in color_key.items()],
        loc="upper center",
        bbox_to_anchor=(0.5, 0.5),
        ncol=len(color_key),
        frameon=False,
        edgecolor="black",
        handlelength=0.7,
        handleheight=0.7,
    )
    logger.info(f"Saving to {args.output_prefix}.(pdf|png)")
    fig.savefig(f"{args.output_prefix}.pdf", bbox_inches="tight", dpi=300)
    fig.savefig(f"{args.output_prefix}.png", bbox_inches="tight", dpi=300)

    return 0



import math
import argparse

import polars as pl

from ..common import BED9P_COLS, add_group_columns


def calculate_qv(args: argparse.Namespace) -> int:
    df_calls = pl.read_csv(
        args.infile,
        separator="\t",
        has_header=False,
        comment_prefix="#",
        columns=list(range(9)),
        schema=dict(BED9P_COLS[0:9]),
        truncate_ragged_lines=True,
    )

    if args.ignore_calls:
        df_calls = df_calls.with_columns(
            name=pl.when(pl.col("name").is_in(args.ignore_calls))
            .then(pl.lit("correct"))
            .otherwise(pl.col("name"))
        )

    # Group regions if not bookended.
    df_calls_grouped = add_group_columns(df_region=df_calls)

    print(
        "#chrom",
        "chromStart",
        "chromEnd",
        "QV",
        "bpError",
        "bpTotal",
        sep="\t",
        file=args.outfile,
    )
    for grp, df_grp in df_calls_grouped.group_by(
        ["#chrom", "group"], maintain_order=True
    ):
        chrom = grp[0]
        min_start = df_grp["minStart"][0]
        max_end = df_grp["maxEnd"][0]

        bp_region = max_end - min_start
        bp_err = df_grp.filter(pl.col("name") != "correct")["length"].sum()

        assert (
            bp_err <= bp_region
        ), f"Length of error region ({bp_err}) exceeds length of region ({bp_region}) for {chrom}"

        try:
            # Same as Inspector's QV metric.
            # https://github.com/ChongLab/Inspector/blob/0e08f882181cc0e0e0fa749cd87fb74a278ea0f0/inspector.py#L184
            # $ python -c "import math,sys; e,t = sys.argv[1:3]; print(-10 * math.log10(int(e) / int(t)))" 1 1670
            #
            # For merqury.
            # $ echo "1 1670" | awk -v k=31 '{print (-10*log(1-(1-$1/$2)^(1/k))/log(10))}'
            #
            # Both increase as numerator drops or denominator increases.
            qv = -10 * math.log10(bp_err / bp_region)
        except ZeroDivisionError:
            qv = 0.0
        except ValueError:
            # No errors in region.
            qv = math.inf

        print(
            chrom,
            min_start,
            max_end,
            qv,
            bp_err,
            bp_region,
            sep="\t",
            file=args.outfile,
        )

    return 0



import random
import logging
import warnings
import matplotlib
import matplotlib.axes
import polars as pl
import matplotlib.pyplot as plt
import matplotlib.patches as ptch

from typing import Any
from collections import OrderedDict
from intervaltree import Interval  # type: ignore[import-untyped]
from matplotlib.collections import PatchCollection
from functools import lru_cache

from .region import ActionOpt, Region
from ..common import minimalize_ax

PLOT_FONT_SIZE = 16
PLOT_HEIGHT = 5
PLOT_WIDTH = 24
PLOT_HEIGHT_SCALE_FACTOR = 2

# No margins.
matplotlib.use("agg")
warnings.filterwarnings("ignore")

# Set text size
matplotlib.rcParams.update({"font.size": PLOT_FONT_SIZE})
plt.rcParams["axes.xmargin"] = 0


@lru_cache
def get_random_color(name: str) -> str:
    # Seed the RNG with the name
    random.seed(name)
    # Then generate rgb and hexcode
    rgb = tuple(random.randint(0, 255) for _ in range(3))
    assert len(rgb) == 3
    # https://stackoverflow.com/a/3380739
    return "#%02x%02x%02x" % rgb


def plot_coverage(
    itv: Interval,
    df_pileup: pl.DataFrame,
    tracks: OrderedDict[str, set[Region]],
    ovl_tracks: OrderedDict[str, set[Region]],
    plot_ylim: float | int = 100,
) -> tuple[plt.Figure, Any]:
    """
    :param itv: Interval to plot
    :type itv: Interval
    :param df_pileup: Pileup dataframe
    :type df_pileup: pl.DataFrame
    :param tracks: Tracks to plot
    :type tracks: OrderedDict[str, set[Region]]
    :param ovl_tracks: Tracks to plot on top of coverage plot.
    :type ovl_tracks: OrderedDict[str, set[Region]]
    :param plot_ylim: Plot y-limit.
    :type plot_ylim: float | int
    :return: Figure and its axes.
    :rtype: tuple[Figure, Any]
    """

    color: str | None
    subplot_patches: dict[str, list[ptch.Rectangle]] = {}
    number_of_overlap_beds = len(tracks.keys()) if tracks else 0
    if tracks:
        fig, axs = plt.subplots(
            # | ---------- |
            # | \/\__/\||/ |
            # | bed_legend |
            # | cov_legend |
            # nrows (n for bed, 1 for the coverage plot, n for bed legends, and 1 for coverage plot legend)
            len(tracks.keys()) + 1 + len(tracks.keys()) + 1,
            # ncols
            1,
            figsize=(
                PLOT_WIDTH,
                PLOT_HEIGHT + (PLOT_HEIGHT_SCALE_FACTOR * number_of_overlap_beds),
            ),
            sharex=True,
            gridspec_kw={
                "height_ratios": [
                    *[0.3 for _ in range(number_of_overlap_beds)],
                    3,
                    *[0.15 for _ in range(number_of_overlap_beds)],
                    0.15,
                ]
            },
            layout="constrained",
        )

        # Last axis with coverage plt.
        ax: matplotlib.axes.Axes = axs[number_of_overlap_beds]
        # Add bed regions.
        for i, (name, regions) in enumerate(tracks.items()):
            # Make axis as minimal as possible.
            bed_axs: matplotlib.axes.Axes = axs[i]
            minimalize_ax(bed_axs, remove_ticks=True)

            bed_axs.set_ylabel(
                name,
                rotation="horizontal",
                ma="right",
                ha="right",
                va="center",
            )

            # Map uniq types to new color if none given.
            patches: list[ptch.Rectangle] = []
            ylim = bed_axs.get_ylim()[1]
            for row in regions:
                # Skip rows not within bounds of df.
                if not itv.overlaps(row.region):
                    continue

                if not row.action or (row.action and row.action.opt != ActionOpt.PLOT):
                    continue

                region_st = max(itv.begin, row.region.begin)
                region_end = min(itv.end, row.region.end)
                width = region_end - region_st
                # Use color provided. Default to random generated ones otherwise.
                if row.action.desc:
                    color = row.action.desc
                elif row.desc:
                    color = get_random_color(row.desc)
                else:
                    raise ValueError(f"Region {row} has no description.")

                rect = ptch.Rectangle(
                    (region_st, 0),
                    width,
                    ylim,
                    linewidth=1,
                    edgecolor=None,
                    facecolor=color,
                    alpha=0.75,
                    label=row.desc,
                )
                patches.append(rect)

            bed_axs.add_collection(PatchCollection(patches, match_original=True))
            subplot_patches[name] = patches
    else:
        fig, axs = plt.subplots(
            2,
            1,
            figsize=(PLOT_WIDTH, PLOT_HEIGHT),
            layout="constrained",
            gridspec_kw={"height_ratios": [3, 0.15]},
        )
        ax = axs[0]

    for typ, markercolor, color, label in (
        ("insertion", "#800080", "#800080", "Insertions"),
        ("deletion", "#f8f4ff", "#000000", "Deletions"),
        ("mismatch", "#ff0000", "#ff0000", "Mismatches"),
        ("cov", "#000000", "#000000", "Coverage"),
    ):
        ax.plot(
            df_pileup["pos"],
            df_pileup[typ],
            marker="o",
            linestyle="None",
            color=color,
            markersize=2,
            markeredgewidth=0.2,
            markerfacecolor=markercolor,
            label=label,
        )

    for track_name, regions in ovl_tracks.items():
        for region in regions:
            if region.name == "correct" and track_name == "Calls":
                continue
            if region.action:
                color = region.action.desc
            elif region.desc:
                color = get_random_color(region.desc)
            else:
                raise ValueError(f"Region {region} has no description.")

            if not region.region.overlaps(itv):
                continue

            region_st = max(itv.begin, region.region.begin)
            region_end = min(itv.end, region.region.end)
            ax.axvspan(
                region_st,
                region_end,
                color=color,
                alpha=0.4,
                label=region.desc,
            )

    # Add legend for coverage plot in separate axis. Deduplicate multiple labels.
    # https://stackoverflow.com/a/36189073
    handles, labels = ax.get_legend_handles_labels()
    labels_handles = dict(zip(labels, handles))
    legend_cov_ax: matplotlib.axes.Axes = axs[-1]
    minimalize_ax(legend_cov_ax, remove_ticks=True)
    legend_cov_ax.legend(
        labels_handles.values(),
        labels_handles.keys(),
        loc="center",
        alignment="left",
        ncols=len(labels),
        borderaxespad=0,
        handlelength=1.0,
        handleheight=1.0,
        fancybox=False,
        frameon=False,
        prop={"size": 12},
        title="Pileup",
    )
    # Add legends for each overlapped bedfile.
    for i, (name, sp_patches) in enumerate(subplot_patches.items()):
        # Remove plot elements from legend ax.
        # Offset by 1 for coverage plot.
        legend_ax: matplotlib.axes.Axes = axs[i + number_of_overlap_beds + 1]
        minimalize_ax(legend_ax, remove_ticks=True)

        # Filter rectangle patches.
        sp_patch_labels = set()
        sp_filtered_patches: list[ptch.Rectangle] = []
        for patch in sp_patches:
            patch.set_edgecolor("black")
            patch_lbl = patch.get_label()
            if patch_lbl in sp_patch_labels:
                continue
            sp_filtered_patches.append(patch)
            sp_patch_labels.add(patch_lbl)

        # Sort patches by label.
        # Special exception for MAPQ since it's a range and cannot sort lexographically. ex. 5-10 and 50-55
        if name == "MAPQ":
            sp_filtered_patches.sort(key=lambda p: int(p.get_label().split("-")[0]))
        elif name == "Self-identity":
            # 99.25%-99.50%
            sp_filtered_patches.sort(
                key=lambda p: float(p.get_label().split("-")[0][0:-1])
            )
        else:
            sp_filtered_patches.sort(key=lambda p: str(p.get_label()))

        legend_ax.legend(
            handles=sp_filtered_patches,
            loc="center",
            alignment="left",
            title=name,
            # At least 1-15 columns in legend.
            ncols=min(max(len(sp_filtered_patches), 1), 15),
            handlelength=1.0,
            handleheight=1.0,
            borderaxespad=0,
            fancybox=False,
            frameon=False,
            prop={"size": 12},
        )

    fig.suptitle("{}:{}-{}\n".format(itv.data, itv.begin, itv.end), fontweight="bold")

    if itv.end < 1_000_000:
        xlabels = [format(label, ",.0f") for label in ax.get_xticks()]
        lab = "bp"
    elif itv.end < 10_000_000:
        xlabels = [format(label / 1000, ",.1f") for label in ax.get_xticks()]
        lab = "kbp"
    else:
        xlabels = [format(label / 1000, ",.1f") for label in ax.get_xticks()]
        lab = "kbp"

    mean_cov = df_pileup["cov"].mean()
    if isinstance(plot_ylim, float):
        assert isinstance(mean_cov, (float, int)), "Invalid mean coverage type."
        plot_ylim = mean_cov * plot_ylim
    elif isinstance(plot_ylim, int):
        plot_ylim = plot_ylim
    else:
        logging.error(
            f"Invalid {plot_ylim} of type {type(plot_ylim)}. Defaulting to 100."
        )
        plot_ylim = 100

    ax.set_ylim(0, plot_ylim)
    ax.set_xlabel("Assembly position ({})".format(lab), fontweight="bold")
    ax.set_ylabel(
        "Sequence\nread\ndepth",
        fontweight="bold",
        rotation="horizontal",
        ma="center",
        ha="right",
        va="center",
    )
    ax.set_xticklabels(xlabels)

    # Hide the right and top spines
    ax.spines["right"].set_visible(False)
    ax.spines["top"].set_visible(False)

    # Only show ticks on the left and bottom spines
    ax.yaxis.set_ticks_position("left")
    ax.xaxis.set_ticks_position("bottom")

    return fig, axs



#!/usr/bin/env python
import os
import time
import pprint
import random
import logging
import tomllib
import tempfile
import argparse
from typing import TextIO
from intervaltree import Interval  # type: ignore[import-untyped]
from collections import OrderedDict, defaultdict
from concurrent.futures import ProcessPoolExecutor, as_completed

import polars as pl
import matplotlib.pyplot as plt

from .plot import plot_coverage
from .io import (
    read_overlay_regions,
    write_output,
    write_bigwig,
    generate_status_from_regions,
    read_identity_breakpoints,
    BED9P_COLS,
)
from .region import (
    Region,
    add_ident_overlay_region,
    add_mapq_overlay_region,
    add_misassemblies_overlay_region,
)
from ..common import DEFAULT_WG_WINDOW, CORRECT_ITEM_RGB

from py_nucflag import run_nucflag_itv, get_regions, get_config_from_preset  # type: ignore[import-untyped]

# Get the logger
logger = logging.getLogger(__name__)


def plot_misassemblies(
    aln: str,
    fasta: str | None,
    itv: tuple[int, int, str],
    ignore_bed: str | None,
    threads: int,
    config: str | None,
    preset: str | None,
    tracks: OrderedDict[str, set[Region]],
    ovl_tracks: OrderedDict[str, set[Region]],
    plot_dir: str | None,
    pileup_dir: str | None,
    add_pileup_data: set[str],
    add_builtin_tracks: set[str],
    ignore_mtypes: list[str],
    overlap_calls: bool,
    ylim: int | float,
    ident_breakpoints: tuple[list[float], list[str]],
) -> pl.DataFrame:
    # This is very broken. It's difficult to sync the logging across processes.
    random.seed(hash(itv))
    wait_time = random.random()
    time.sleep(wait_time)

    # Detect misassemblies and return object with cov pileup and regions.
    res = run_nucflag_itv(
        aln,
        itv=itv,
        fasta=fasta,
        ignore_bed=ignore_bed,
        threads=threads,
        cfg=config,
        preset=preset,
    )
    st, end, ctg = itv
    ctg_coords = f"{ctg}:{st}-{end}"
    ctg_coords_filesafe = f"{ctg}_{st}-{end}"

    # Ignore mtypes from results.
    if ignore_mtypes:
        # Must be sorted and from one chrom.
        # Not writable.
        regions = (
            res.regions.with_columns(
                name=pl.when(pl.col("name").is_in(ignore_mtypes))
                .then(pl.lit("correct"))
                .otherwise(pl.col("name")),
                itemRgb=pl.when(pl.col("name").is_in(ignore_mtypes))
                .then(pl.lit(CORRECT_ITEM_RGB))
                .otherwise(pl.col("itemRgb")),
            )
            .with_columns(grp=pl.col("name").rle_id())
            .group_by(["grp"])
            .agg(
                pl.col("#chrom").first(),
                pl.col("chromStart").min(),
                pl.col("chromEnd").max(),
                pl.col("name").first(),
                # May change as result of filtering.
                pl.col("score").median().round().cast(pl.UInt64),
                pl.col("strand").first(),
                pl.col("thickStart").min(),
                pl.col("thickEnd").max(),
                pl.col("itemRgb").first(),
                pl.col("zscore").max(),
                pl.col("af").max(),
            )
            .drop("grp")
        )
    else:
        regions = res.regions

    # Plot contig.
    if plot_dir and isinstance(res.pileup, pl.DataFrame):
        if "ident" in add_builtin_tracks:
            logger.info(f"Adding local self-identity track for {ctg_coords}.")
            tracks["Self-identity"] = set(
                add_ident_overlay_region(
                    ctg, res.pileup.select("pos", "bin", "bin_ident"), ident_breakpoints
                )
            )

        if "mapq" in add_builtin_tracks:
            logger.info(f"Adding mapq track for {ctg_coords}.")
            tracks["MAPQ"] = set(
                add_mapq_overlay_region(ctg, res.pileup.select("pos", "mapq"))
            )

        if overlap_calls:
            ovl_tracks["Calls"] = set(add_misassemblies_overlay_region(regions))
        else:
            tracks["Calls"] = set(add_misassemblies_overlay_region(regions))

        logger.info(f"Plotting {ctg_coords}.")
        fig, _ = plot_coverage(
            itv=Interval(st, end, ctg),
            df_pileup=res.pileup,
            tracks=tracks,
            ovl_tracks=ovl_tracks,
            plot_ylim=ylim,
        )

        output_plot = os.path.join(plot_dir, f"{ctg_coords_filesafe}.png")
        logger.info(f"Saving plot to {output_plot}.")
        plt.savefig(output_plot, dpi=600, bbox_inches="tight")
        plt.close(fig)

    # Output coverage.
    if pileup_dir and add_pileup_data:
        logger.info(f"Saving pileup data for {ctg_coords} to {pileup_dir}.")
        write_bigwig(
            res.pileup,
            chrom_lengths=f"{fasta}.fai",
            columns=list(add_pileup_data),
            output_dir=os.path.join(pileup_dir),
        )

    return regions


def call_misassemblies(args: argparse.Namespace) -> int:
    if args.output_plot_dir:
        os.makedirs(args.output_plot_dir, exist_ok=True)
    if args.output_pileup_dir:
        os.makedirs(args.output_pileup_dir, exist_ok=True)

    if not (isinstance(args.ylim, float) or isinstance(args.ylim, int)):
        raise ValueError(f"y-axis limit must be float or int. {args.ylim}")

    # Load ignored regions.
    ignore_bed = None
    tmpfile_ignore_bed = tempfile.NamedTemporaryFile("wt")
    if args.ignore_regions:
        logger.info(f"Ignoring region(s) from {args.ignore_regions}.")
        with open(args.ignore_regions, "rt") as fh:
            for line in fh:
                tmpfile_ignore_bed.write(line)
        ignore_bed = tmpfile_ignore_bed

    # Load additional regions to overlay and ignore.
    if args.tracks:
        tracks: defaultdict[str, OrderedDict[str, set[Region]]] = defaultdict(
            OrderedDict
        )
        tracks = read_overlay_regions(args.tracks)
        logger.info(f"Adding {len(args.tracks)} bedfile(s).")
    else:
        tracks = defaultdict(OrderedDict)

    if args.overlap_tracks:
        ovl_tracks: defaultdict[str, OrderedDict[str, set[Region]]] = defaultdict(
            OrderedDict
        )
        ovl_tracks = read_overlay_regions(args.overlap_tracks)
        logger.info(f"Overlapping {len(args.overlap_tracks)} bedfile(s).")
    else:
        ovl_tracks = defaultdict(OrderedDict)

    if args.config:
        with open(args.config, "rb") as fh:
            cfg = tomllib.load(fh)
        cfg_general: dict = cfg.get("general", {})
        window = cfg_general.get("bp_wg_window", DEFAULT_WG_WINDOW)
    else:
        window = DEFAULT_WG_WINDOW

    # Print config to stderr.
    config_str = get_config_from_preset(args.preset, args.config)
    cfg = tomllib.loads(config_str)
    logger.info(f"Using config:\n{pprint.pformat(cfg, underscore_numbers=True)}")

    regions: list[tuple[int, int, str]] = get_regions(
        aln=args.infile,
        bed=args.input_regions.name if args.input_regions else None,
        window=window,
    )
    if not regions:
        logger.error(
            f"No valid regions in {args.input_regions.name if args.input_regions else args.infile}."
        )
        return 1

    # Load identity breakpoints
    ident_breakpoints = read_identity_breakpoints(args.ident_breakpoints)

    # Ignore types.
    ignore_mtypes = []
    if args.ignore_mtypes:
        ignore_mtypes = args.ignore_mtypes
        logger.info(f"Ignoring {ignore_mtypes} from output bed and plots.")

    # Tracks to add that are builtin to nucflag.
    # mapq/ident/...
    added_builtin_tracks = (
        set(args.add_builtin_tracks) if args.add_builtin_tracks else set()
    )
    added_pileup_data = set(args.add_pileup_data) if args.add_pileup_data else set()
    dfs_regions: list[pl.DataFrame] = []
    logger.info(
        f"Generating outputs with {args.processes} processes with nucflag running with {args.threads} threads."
    )
    save_res = args.output_regions.name == "<stdout>" and args.output_status
    all_args = [
        [
            args.infile,
            args.fasta,
            rgn,
            ignore_bed.name if ignore_bed else None,
            args.threads,
            args.config,
            args.preset,
            tracks.get(rgn[2], OrderedDict()),
            ovl_tracks.get(rgn[2], OrderedDict()),
            args.output_plot_dir,
            args.output_pileup_dir,
            added_pileup_data,
            added_builtin_tracks,
            ignore_mtypes,
            args.overlap_calls,
            args.ylim,
            ident_breakpoints,
        ]
        for rgn in regions
    ]
    if args.processes == 1:
        for a in all_args:
            res = plot_misassemblies(*a)
            # Write to file as soon as done.
            res.write_csv(args.output_regions, include_header=False, separator="\t")
            dfs_regions.append(res)
    else:
        with ProcessPoolExecutor(
            max_workers=args.processes, max_tasks_per_child=1
        ) as pool:
            futures = {pool.submit(plot_misassemblies, *a): a[2] for a in all_args}
            for future in as_completed(futures):
                if future.exception():
                    st, end, chrom = futures[future]
                    raise RuntimeError(
                        f"Failed to write output for {chrom}:{st}-{end} ({future.exception()})"
                    )
                df_res = future.result()
                # Write to file as soon as done.
                df_res.write_csv(
                    args.output_regions, include_header=False, separator="\t"
                )
                # But if writing to stdout, cannot retrieve later for status so save.
                if save_res:
                    dfs_regions.append(df_res)

    if save_res:
        output_regions: list[pl.DataFrame] | TextIO = dfs_regions
    else:
        output_regions = args.output_regions

    # Remove tempfile of ignored regions.
    tmpfile_ignore_bed.close()

    logger.info(f"Writing region BED file to {args.output_regions.name}.")
    write_output(
        output_regions,
        args.output_status,
    )
    logger.info("Done!")
    return 0


def create_status(args: argparse.Namespace) -> int:
    df_regions = pl.read_csv(
        args.infile,
        separator="\t",
        has_header=False,
        comment_prefix="#",
        columns=list(range(9)),
        schema=dict(BED9P_COLS[0:9]),
        truncate_ragged_lines=True,
    )
    if df_regions.is_empty():
        raise ValueError(f"No regions to generate status for {args.infile}")

    df_status = generate_status_from_regions(df_regions)
    df_status.write_csv(file=args.outfile, include_header=True, separator="\t")
    return 0



import polars as pl
from matplotlib.axes import Axes

BED9P_COLS = [
    ("#chrom", pl.String),
    ("chromStart", pl.UInt64),
    ("chromEnd", pl.UInt64),
    ("name", pl.String),
    ("score", pl.UInt64),
    ("strand", pl.String),
    ("thickStart", pl.UInt64),
    ("thickEnd", pl.UInt64),
    ("itemRgb", pl.String),
    ("zscore", pl.Float32),
    ("af", pl.Float32),
]

STATUSES = (
    "correct",
    "insertion",
    "deletion",
    "softclip",
    "low_quality",
    "het_or_mismap",
    "collapse",
    "misjoin",
    "mismatch",
    "false_dup",
    "homopolymer",
    "dinucleotide",
    "simple_repeat",
    "other_repeat",
    "scaffold",
)
PRESETS = ("ont_r9", "ont_r10", "hifi")
DEFAULT_WG_WINDOW = 5_000_000
CORRECT_ITEM_RGB = "206,206,206"


def minimalize_ax(ax: Axes, *, remove_ticks: bool = False) -> None:
    for spine in ["left", "right", "bottom", "top"]:
        ax.spines[spine].set_visible(False)
    if remove_ticks:
        ax.tick_params(
            axis="both",
            left=False,
            top=False,
            right=False,
            bottom=False,
            labelleft=False,
            labeltop=False,
            labelright=False,
            labelbottom=False,
        )


def add_group_columns(df_region: pl.DataFrame) -> pl.DataFrame:
    return (
        df_region.sort(by=["#chrom", "chromStart"])
        .with_columns(
            group=(pl.col("chromEnd") != pl.col("chromStart").shift(-1))
            .fill_null(False)
            .rle_id()
            .over("#chrom"),
        )
        .with_columns(
            group=pl.when(pl.col("group") % 2 != 0)
            .then(pl.col("group") - 1)
            .otherwise(pl.col("group"))
            .over("#chrom")
        )
        .with_columns(
            length=pl.col("chromEnd") - pl.col("chromStart"),
            minStart=pl.col("chromStart").min().over(["#chrom", "group"]),
            maxEnd=pl.col("chromEnd").max().over(["#chrom", "group"]),
        )
    )



import sys
import argparse

from typing import TYPE_CHECKING, Any


from ..common import STATUSES

if TYPE_CHECKING:
    SubArgumentParser = argparse._SubParsersAction[argparse.ArgumentParser]
else:
    SubArgumentParser = Any


def add_qv_cli(parser: SubArgumentParser) -> None:
    ap = parser.add_parser(
        "qv",
        description="Calculate QV from NucFlag misassemblies based on formula: -10 * log10(bp_err / bp_region)",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    ap.add_argument(
        "-i",
        "--infile",
        required=True,
        type=argparse.FileType("rb"),
        help="Input NucFlag misassembly calls as BED9.",
    )
    ap.add_argument(
        "-o",
        "--outfile",
        default=sys.stdout,
        type=argparse.FileType("wt"),
        help="Output BED4 file with QV as 4th column.",
    )
    ap.add_argument(
        "-c",
        "--ignore-calls",
        help="Ignore specific calls in QV calculation.",
        nargs="*",
        default=["scaffold"],
        type=str,
        choices=[s for s in STATUSES if s != "correct"],
    )

    return None



import argparse

from typing import TYPE_CHECKING, Any


if TYPE_CHECKING:
    SubArgumentParser = argparse._SubParsersAction[argparse.ArgumentParser]
else:
    SubArgumentParser = Any


def add_ideogram_cli(parser: SubArgumentParser) -> None:
    ap = parser.add_parser(
        "ideogram",
        description="Generate ideogram from NucFlag misassemblies",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    ap.add_argument(
        "-i",
        "--infile",
        required=True,
        type=argparse.FileType("rb"),
        help="Input whole-genome NucFlag misassembly calls as BED9.",
    )
    ap.add_argument(
        "-c",
        "--cytobands",
        default=None,
        help="Cytoband file. Format: [chrom, start, end, name, btype]. See https://github.com/Balthasar-eu/pyideogram.",
    )
    ap.add_argument(
        "-l",
        "--filter_length",
        type=int,
        default=10_000_000,
        help="Only plot contigs at least this long.",
    )
    ap.add_argument(
        "-t",
        "--track_height",
        default=2.0,
        type=float,
        help="Height of individual track.",
    )
    ap.add_argument(
        "-o", "--output_prefix", default="ideogram", help="Output file prefix."
    )

    return None



import logging
import argparse
import polars as pl
import pyideogram as pyid
import matplotlib.pyplot as plt
import matplotlib.patheffects as pe

from typing import Any
from matplotlib.axes import Axes
from matplotlib import cbook
from matplotlib.colors import rgb2hex
from matplotlib.patches import FancyBboxPatch, Patch
from pyideogram.matplotlib_extension import SideRound

from ..common import BED9P_COLS, minimalize_ax

BAND_COLORS = pyid.BANDCOL | {"none": (1.0, 1.0, 1.0)}
LBL_KWARGS = dict(rotation=0, ha="right", va="center")


logger = logging.getLogger(__name__)


def draw_ideogramh_no_cytobands(
    ax: Axes, chrom: str, length: int, textkwargs: dict[str, Any] = {}, **ideokwargs
):
    ideokwargs = cbook.normalize_kwargs(ideokwargs, Patch)
    ideokwargs = {"edgecolor": "k"} | ideokwargs
    barpatches = ax.barh(
        [chrom],
        length,
        1,
        0,
        color=BAND_COLORS["gneg"],
        **ideokwargs,
    )
    cornerref = (0, 3, 1, 2)
    for val in [
        "edgecolor",
        "linewidth",
        "hatch",
        "xerr",
        "yerr",
        "error_kw",
        "ecolor",
        "capsize",
        "orientation",
    ]:
        ideokwargs.pop(val, None)

    fs = 5
    textkwargs = {
        "path_effects": [
            pe.withStroke(linewidth=fs / 3, foreground="white", alpha=0.8)
        ],
        "fontsize": fs,
        "va": "center_baseline",
        "ha": "center",
        "weight": "bold",
        "clip_on": True,
    } | textkwargs

    figW, figH = ax.get_figure().get_size_inches()
    _, _, w, h = ax.get_position().bounds
    disp_ratio = (figW * w) / (figH * h)
    ratio = ax.get_data_ratio() * disp_ratio
    # Get patch
    patch = barpatches[0]
    bb = patch.get_bbox()
    patch.remove()
    Box = SideRound(
        xround=min([((0.55 * 0.9) / ratio) / bb.width, 2]),
        yround=0.9,
        corners=cornerref,
    )
    fp = FancyBboxPatch(
        (bb.xmin, bb.ymin),
        abs(bb.width),
        abs(bb.height),
        boxstyle=Box,
        ec=patch.get_edgecolor(),
        fc=patch.get_facecolor(),
        linewidth=patch.get_linewidth(),
        hatch=patch.get_hatch(),
        label=patch.get_label(),
    )
    fp._internal_update(ideokwargs)
    ax.add_patch(fp)


def create_ideogram(args: argparse.Namespace) -> int:
    df_calls = pl.read_csv(
        args.infile,
        separator="\t",
        has_header=False,
        comment_prefix="#",
        columns=list(range(9)),
        schema=dict(BED9P_COLS[0:9]),
        truncate_ragged_lines=True,
    ).with_columns(length=pl.col("chromEnd") - pl.col("chromStart"))

    df_fai = df_calls.group_by(["#chrom"]).agg(
        length=pl.col("chromEnd").max() - pl.col("chromStart").min()
    )
    fai_map = dict(df_fai.iter_rows())
    if args.cytobands:
        cytobands = pyid.dataloader.load_cytobands(args.cytobands)
    else:
        cytobands = None

    color_key = {
        name: rgb2hex([int(e) / 255.0 for e in itemRgb.split(",")])
        if not itemRgb.startswith("#")
        else itemRgb
        for name, itemRgb in df_calls.select("name", "itemRgb").unique().iter_rows()
    }
    max_length = df_fai["length"].max()
    chrom_names = (
        df_fai.filter(pl.col("length") > args.filter_length)
        .sort(by="length", descending=True)["#chrom"]
        .unique(maintain_order=True)
    )
    logger.info(f"Plotting {len(chrom_names)} chromosomes.")
    if chrom_names.is_empty():
        return 1

    # chrom, calls, spacer
    base_height_ratios = [1.0, 0.66]
    width_ratios = [0.1, 0.9]
    num_tracks = len(base_height_ratios)
    height_ratios = base_height_ratios * len(chrom_names)
    # Add extra for legend
    height_ratios.append(3.0)

    fig, axes = plt.subplots(
        ncols=len(width_ratios),
        # 1 additional track for legend
        nrows=len(chrom_names) * num_tracks + 1,
        figsize=(20, len(chrom_names) * args.track_height),
        height_ratios=height_ratios,
        width_ratios=width_ratios,
    )
    fig.subplots_adjust(wspace=0.02)

    ax_col_idx_stats = 0
    ax_col_idx_chrom = 1

    for chrom_name_idx, chrom_name in enumerate(chrom_names):
        df_chrom_calls = df_calls.filter(pl.col("#chrom") == chrom_name)
        chrom_length = fai_map[chrom_name]

        logger.info(
            f"On contig #{chrom_name_idx + 1} {chrom_name} ({chrom_length // 1_000_000} Mbp) ..."
        )
        ax_row_idx_chrom = chrom_name_idx * num_tracks
        ax_row_indices_tracks = range(
            ax_row_idx_chrom + 1, ax_row_idx_chrom + num_tracks
        )
        ax_chrom: Axes = axes[ax_row_idx_chrom, ax_col_idx_chrom]
        ax_chrom_stats: Axes = axes[ax_row_idx_chrom, ax_col_idx_stats]

        ax_chrom.xaxis.set_tick_params(which="both", length=0, labelleft=False)
        ax_chrom.yaxis.set_tick_params(which="both", length=0)
        # pyideogram removes xyticks
        minimalize_ax(ax_chrom)
        minimalize_ax(ax_chrom_stats, remove_ticks=True)

        for ax_row_idx in ax_row_indices_tracks:
            ax_track: Axes = axes[ax_row_idx, ax_col_idx_chrom]
            ax_stats: Axes = axes[ax_row_idx, ax_col_idx_stats]

            ax_track.set_xlim(0, max_length)
            ax_track.set_ylim(0, 1)
            minimalize_ax(ax_track, remove_ticks=True)
            minimalize_ax(ax_stats, remove_ticks=True)

            # Write stats
            type_counts = dict(
                df_chrom_calls.group_by(["name"])
                .agg(length=pl.col("length").sum())
                .iter_rows()
            )
            ax_stats.pie(
                type_counts.values(),
                colors=[color_key[typ] for typ in type_counts.keys()],
                radius=1,
            )

            # Write regions
            for row in df_chrom_calls.iter_rows(named=True):
                color = color_key[row["name"]]
                ax_track.axvspan(
                    xmin=row["chromStart"], xmax=row["chromEnd"], color=color
                )

        ax_chrom.set_xlim(0, max_length)
        if cytobands:
            pyid.ideogramh(
                chrom=chrom_name,
                bands=cytobands,
                ax=ax_chrom,
                color=BAND_COLORS,
                label="",
            )
        else:
            draw_ideogramh_no_cytobands(
                ax=ax_chrom, chrom=chrom_name, length=chrom_length
            )

        ax_chrom.set_yticks([], [])
        ax_chrom.set_ylabel(chrom_name, **LBL_KWARGS)

    # Add legend.
    ax_legend: Axes = axes[len(chrom_names) * num_tracks, ax_col_idx_chrom]
    minimalize_ax(ax_legend, remove_ticks=True)
    minimalize_ax(
        axes[len(chrom_names) * num_tracks, ax_col_idx_stats], remove_ticks=True
    )
    ax_legend.legend(
        handles=[Patch(facecolor=color, label=lbl) for lbl, color in color_key.items()],
        bbox_to_anchor=(0.5, 0.5),
        loc="center",
        ncol=6,
        frameon=False,
        edgecolor="black",
        handlelength=0.7,
        handleheight=0.7,
    )
    # Reduce white space between haps
    logger.info(f"Saving to {args.output_prefix}.(pdf|png)")
    fig.savefig(f"{args.output_prefix}.pdf", bbox_inches="tight", dpi=600)
    fig.savefig(f"{args.output_prefix}.png", bbox_inches="tight", dpi=600)

    return 0



from enum import Enum, auto
from typing import Generator, NamedTuple

import polars as pl
import matplotlib.colors

from bisect import bisect
from intervaltree import Interval  # type: ignore[import-untyped]


class ActionOpt(Enum):
    IGNORE = auto()
    PLOT = auto()
    NOOP = auto()


class Action(NamedTuple):
    opt: ActionOpt
    desc: str | None

    def as_str(self) -> str:
        action = f"{self.opt}"
        if self.desc:
            action = f"{action}:{self.desc}"
        return action


class Region(NamedTuple):
    name: str
    region: Interval
    desc: str | None
    action: Action | None

    def as_tsv(self) -> str:
        action_str = self.action.as_str() if self.action else "None"
        return f"{self.name}\t{self.region.begin}\t{self.region.end}\t{self.desc}\t{action_str}"


DF_MAPQ_COLORS = pl.DataFrame(
    {
        "mapq_rng": [
            "0",
            "1-5",
            "5-10",
            "10-15",
            "15-20",
            "20-25",
            "25-30",
            "30-35",
            "35-40",
            "40-45",
            "45-50",
            "50-55",
        ],
        "mapq_color": [
            "#666666",
            "#8f59a7",
            "#5954a8",
            "#01aef3",
            "#04b99e",
            "#8bc83b",
            "#cdde25",
            "#fff600",
            "#ffc309",
            "#fa931a",
            "#f8631f",
            "#f21821",
        ],
    }
)


def add_mapq_overlay_region(
    name: str, df: pl.DataFrame
) -> Generator[Region, None, None]:
    regions = (
        df.with_columns(mapq_grp=pl.col("mapq").rle_id())
        .group_by(["mapq_grp"])
        .agg(
            st=pl.col("pos").min(), end=pl.col("pos").max(), mapq=pl.col("mapq").first()
        )
        .with_columns(mapq_lower_bound=(pl.col("mapq") // 5) * 5)
        .with_columns(
            mapq_rng=pl.when(pl.col("mapq_lower_bound") == 60)
            .then(pl.lit("55-60"))
            # 0
            .when(pl.col("mapq") == 0)
            .then(pl.lit("0"))
            # 0-5 to 1-5
            .when(pl.col("mapq_lower_bound") == 0)
            .then(pl.lit("1-5"))
            # 5-60
            .otherwise(
                pl.col("mapq_lower_bound").cast(pl.String)
                + "-"
                + (pl.col("mapq_lower_bound") + 5).clip(0, 60).cast(pl.String)
            )
        )
        .join(DF_MAPQ_COLORS, on=["mapq_rng"], how="left")
        .with_columns(pl.col("mapq_color").fill_null(pl.lit("#FF8DA1")))
        .select("st", "end", "mapq_rng", "mapq_color")
    )
    for st, end, mapq_rng, mapq_color in regions.iter_rows():
        yield Region(
            name,
            Interval(st, end),
            desc=mapq_rng,
            action=Action(ActionOpt.PLOT, mapq_color),
        )


def add_ident_overlay_region(
    name: str, df: pl.DataFrame, breakpoint_colors: tuple[list[float], list[str]]
) -> Generator[Region, None, None]:
    breakpoints, colors = breakpoint_colors
    regions = (
        df.with_columns(bin_grp=pl.col("bin").rle_id())
        .group_by(["bin_grp"])
        .agg(
            st=pl.col("pos").min(),
            end=pl.col("pos").max(),
            bin_ident=pl.col("bin_ident").mean(),
        )
        .select("st", "end", "bin_ident")
    )
    for st, end, bin_ident in regions.iter_rows():
        if bin_ident == 0.0:
            continue
        try:
            idx_end = bisect(breakpoints, bin_ident)
            ident_end = breakpoints[idx_end]
            if idx_end == 0:
                ident_st = 0.0
            else:
                ident_st = breakpoints[idx_end - 1]
                color = colors[idx_end - 1]
            desc = f"{ident_st}%-{ident_end}%"
        except IndexError:
            ident_end = breakpoints[-1]
            color = colors[-1]
            desc = str(ident_end)

        yield Region(
            name,
            Interval(st, end),
            desc=desc,
            action=Action(ActionOpt.PLOT, color),
        )


def add_misassemblies_overlay_region(df: pl.DataFrame) -> Generator[Region, None, None]:
    df = df.with_columns(
        itemRgb=pl.when(pl.col("itemRgb").str.contains(","))
        .then(
            pl.col("itemRgb")
            .str.split(",")
            .list.eval(pl.element().cast(pl.Float32) / 255.0)
            .map_elements(lambda x: matplotlib.colors.to_hex(x), return_dtype=pl.String)
        )
        .otherwise(pl.col("itemRgb"))
    ).select("chromStart", "chromEnd", "name", "itemRgb")
    for row in df.iter_rows(named=True):
        yield Region(
            row["name"],
            Interval(row["chromStart"], row["chromEnd"]),
            desc=row["name"],
            action=Action(ActionOpt.PLOT, row["itemRgb"]),
        )



#!/usr/bin/env python
import sys
import time
import logging
import argparse

from importlib.metadata import version

from .call import add_call_cli, add_status_cli, call_misassemblies, create_status
from .ideogram import add_ideogram_cli, create_ideogram
from .breakdown import add_breakdown_cli, create_breakdown_plot
from .qv import add_qv_cli, calculate_qv
from .consensus import add_consensus_cli, get_consensus_calls
from .config import add_config_cli, get_config

# Configure logging format to match rs-nucflag
# Set UTC
logging.Formatter.converter = time.gmtime
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s.%(msecs)03dZ \033[32m%(levelname)s\033[0m  [py_nucflag::%(name)s] %(message)s",
    datefmt="%Y-%m-%dT%H:%M:%S",
)

# Create the logger
logger = logging.getLogger(__name__)


def main() -> int:
    ap = argparse.ArgumentParser(
        description="Use per-base read coverage to classify/plot misassemblies.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    ap.add_argument("-v", "--version", action="version", version=version("nucflag"))

    sub_ap = ap.add_subparsers(dest="cmd")
    add_call_cli(sub_ap)
    add_status_cli(sub_ap)
    add_ideogram_cli(sub_ap)
    add_breakdown_cli(sub_ap)
    add_qv_cli(sub_ap)
    add_consensus_cli(sub_ap)
    add_config_cli(sub_ap)

    args = ap.parse_args()

    if args.cmd == "call":
        return call_misassemblies(args)
    elif args.cmd == "status":
        return create_status(args)
    elif args.cmd == "ideogram":
        return create_ideogram(args)
    elif args.cmd == "breakdown":
        return create_breakdown_plot(args)
    elif args.cmd == "qv":
        return calculate_qv(args)
    elif args.cmd == "consensus":
        return get_consensus_calls(args)
    elif args.cmd == "config":
        return get_config(args)
    else:
        ap.print_help(sys.stderr)
        return 1


if __name__ == "__main__":
    raise SystemExit(main())



import argparse
import statistics

import matplotlib.colors

import polars as pl
import matplotlib.pyplot as plt

from typing import TextIO, Callable
from collections import defaultdict

from intervaltree import Interval, IntervalTree

from .constants import GOOD_REGIONS
from ..common import BED9P_COLS


def get_consensus_calls(args: argparse.Namespace) -> int:
    files: list[TextIO] = args.infiles

    ignore_calls = set(GOOD_REGIONS)
    if args.ignore_calls:
        ignore_calls.update(args.ignore_calls)

    assert len(files) > 1, f"Only one file ({[file.name for file in files]}) provided."

    # {chrom: IntervalTree[Interval[start, end, name]]}
    itree_calls: defaultdict[str, IntervalTree] = defaultdict(IntervalTree)
    for file in files:
        df_call = pl.read_csv(
            file,
            separator="\t",
            has_header=False,
            comment_prefix="#",
            columns=list(range(4)),
            schema=dict(BED9P_COLS[0:4]),
            truncate_ragged_lines=True,
        )
        df_call = df_call.filter(~pl.col("name").is_in(ignore_calls))
        for call in df_call.iter_rows(named=True):
            itree_calls[call["#chrom"]].add(
                Interval(
                    max(call["chromStart"] - args.dst, 0),
                    call["chromEnd"] + args.dst,
                    # Add filename so unique if name is same.
                    (call["name"], file.name),
                )
            )

    if args.perc_ovl:
        assert args.perc_ovl < 1.0, f"Percent overlap ({args.perc_ovl}) exceeds 100%"

    fn_ovl_check: Callable[[float, float], tuple[bool, float]] | None
    if args.perc_ovl and args.type_ovl == "A":

        def fn_ovl_check(a_b_ovl_frac, b_a_ovl_frac):
            return (b_a_ovl_frac > args.perc_ovl, b_a_ovl_frac)

    elif args.perc_ovl and args.type_ovl == "B":

        def fn_ovl_check(a_b_ovl_frac, b_a_ovl_frac):
            return (
                a_b_ovl_frac > args.perc_ovl,
                a_b_ovl_frac,
            )

    elif args.perc_ovl and args.type_ovl == "A_B":

        def fn_ovl_check(a_b_ovl_frac, b_a_ovl_frac):
            return (
                a_b_ovl_frac > args.perc_ovl and b_a_ovl_frac > args.perc_ovl,
                (a_b_ovl_frac + b_a_ovl_frac) / 2.0,
            )
    else:
        fn_ovl_check = None

    # https://stackoverflow.com/a/43009963
    cmap = plt.cm.get_cmap("rainbow")
    if args.perc_ovl:
        norm = matplotlib.colors.Normalize(vmin=0, vmax=100)
    else:
        norm = matplotlib.colors.Normalize(vmin=1, vmax=len(files))

    # Reduce overlapping intervals
    def data_reducer(
        curr: tuple[set[str], int, str], new: tuple[set[str], int, str]
    ) -> tuple[set[str], int, str]:
        return (curr[0].union(new[0]), max(curr[1], new[1]), curr[2])

    for chrom, itrees in itree_calls.items():
        itv: Interval
        ovl: set[Interval]
        all_itvs = IntervalTree()

        for itv in itrees.iter():
            # a - itv
            ovl = itrees.overlap(itv)

            # Check overlaps to ensure meets overlap criteria.
            if ovl and args.perc_ovl and fn_ovl_check:
                itv_len = itv.length()
                new_ovl = set()
                for ovl_itv in ovl:
                    # b - itv_ovl
                    ovl_itv_len = ovl_itv.length()
                    a_b_ovl_frac = itv_len / ovl_itv_len
                    b_a_ovl_frac = ovl_itv_len / itv_len
                    is_valid, ovl_frac = fn_ovl_check(a_b_ovl_frac, b_a_ovl_frac)
                    if is_valid:
                        # (name, filename, ovl_frac)
                        mdata: tuple[str, str, str] = (*ovl_itv.data, ovl_frac)
                        valid_itv = Interval(ovl_itv.begin, ovl_itv.end, mdata)
                        new_ovl.add(valid_itv)

                ovl = new_ovl

            # Count by total file names so 2 <= valid_n <= n_files
            valid_ovls = set(oitv.data[1] for oitv in ovl)
            valid_n = len(valid_ovls) >= args.total_number_ovl
            if not valid_n:
                continue

            # Color by average overlap percent
            # Or number of overlaps
            if args.perc_ovl:
                ovl_avg = int(statistics.mean([oitv.data[2] for oitv in ovl]) * 100)
            else:
                ovl_avg = len(valid_ovls)

            ovl_names = set(oitv.data[0] for oitv in ovl)
            color = cmap(norm(ovl_avg))
            rgb = ",".join(str(int(c * 255)) for c in color[0:3])
            all_itvs.add(Interval(itv.begin, itv.end, (ovl_names, ovl_avg, rgb)))

        # Merge overlaps
        all_itvs.merge_overlaps(
            data_reducer=data_reducer,
        )
        for itv in sorted(all_itvs):
            ovl_names, ovl_avg, rgb = itv.data
            # Don't modify zero start if dst non-zero.
            st = str(0 if itv.begin == 0 else itv.begin + args.dst)
            end = str(itv.end - args.dst)
            row = (
                chrom,
                st,
                end,
                ",".join(sorted(ovl_names)),
                ".",
                str(ovl_avg),
                st,
                end,
                ",".join(str(int(c * 255)) for c in color[0:3]),
            )
            print("\t".join(row), file=args.outfile)

    return 0
